/* Copyright (c) 2009 Ondrej Donek.
   See the file LICENSE.txt for licensing information. */

// Namespace
if(typeof(extensions.tbnotes.db) == "undefined") extensions.tbnotes.db = {};

/**
 * This serves all about storage. Should not be used by end developer. They should 
 * rather use data objects prototypes instead (e.g. extensions.tbnotes.data.*Prototype).
 * Each of these data objects provide same simple interface for storing data.
 * 
 * @link http://developer.mozilla.org/en/Extensions/Thunderbird/HowTos/Common_Thunderbird_Extension_Techniques/Use_SQLite
 */
(function() {
  // 'Private' properties
  var __initialized = false;
  var __dirSrv = null;
  var __dbFile = null;
  var __dbSrv = null;
  var __dbCon = null;
  // Object describing our database schema
  var __dbSchema = {
    tables: {
      tbn_notes:
        "id INTEGER PRIMARY KEY, \
         target_type TEXT NOT NULL DEFAULT '', \
         target_key TEXT NOT NULL DEFAULT '', \
         category_id INTEGER NOT NULL, \
         created TEXT NOT NULL, \
         updated TEXT NOT NULL, \
         title TEXT NOT NULL, \
         body TEXT, \
         deleted INTEGER DEFAULT 0 ",
      tbn_categories:
        "id INTEGER PRIMARY KEY, \
         parent_id INTEGER, \
         name TEXT NOT NULL, \
         description TEXT ",
      tbn_links:
        "id INTEGER PRIMARY KEY, \
         note_id INTEGER, \
         url TEXT "
    }
  };
  // Holds array of queries for updating db (used in development versions)
  var __dbSchemaUpdates = {
    // from 0.1.1 to 0.1.5
    updateNotes: "ALTER TABLE tbn_notes ADD COLUMN target_name TEXT NOT NULL DEFAULT '' " ,
    updateCategories: "ALTER TABLE tbn_categories ADD COLUMN color TEXT NOT NULL DEFAULT '' ",
		// from 0.1.5 to 0.2
		updateLinks: "ALTER TABLE tbn_links ADD COLUMN title TEXT NOT NULL DEFAULT '' ",
		updateTargetTypes1: "UPDATE `tbn_notes` SET `target_type` = '' WHERE  `target_type` = 'none' ",
		updateTargetTypes2: "UPDATE `tbn_notes` SET `target_type` = 'mailmessage' WHERE  `target_type` = 'message' " 
  };
  
  /**
   * Returns TRUE if tbNotes'db part was successfully initialized.
   * 
   * @return Boolean
   */
  this.isInitialized = function() {
    // And what about `__dbCon.connectionReady`
    return __initialized;
  };
  
  /**
   * Initialize database and establish the connection
   * 
   * @returns boolean
   */
  this.init = function() 
  {
    try {
      __dirSrv = CC["@mozilla.org/file/directory_service;1"].getService(CI.nsIProperties);
      __dbFile = __dirSrv.get("ProfD", CI.nsIFile);  
      __dbFile.append("tbnotes.sqlite");
      
      // Initialize db service and open connection
      __dbSrv = CC["@mozilla.org/storage/service;1"].getService(CI.mozIStorageService);
      __dbCon = __dbSrv.openDatabase(__dbFile);
      
      // Check if all tables exist
      for(let name in __dbSchema.tables) {
        if(!__dbCon.tableExists(name)) {
          __dbCon.createTable(name, __dbSchema.tables[name]);
        }
      }
      
      // Runs update commands
      for(let name in __dbSchemaUpdates) {
        try { 
          __dbCon.executeSimpleSQL(__dbSchemaUpdates[name]);
        } catch(e) {
          CU.reportError("Error occured when running update query '" + name + "'. " +
												 "Probably database is already updated.");
        }
      }
    } catch(e) {
      CU.reportError(e);
      
      // TODO Use stringbundles and normal propmt service
      alert("Error occured during initialization database connection.\n" + 
            "tbNotes'll not work correctly.");
      
      return false;
    }
    
    return false;
  };
  
  // ========================================================================
  // Notes
	
	/**
	 * Returns note by its id. Otherwise returns NULL.
	 * 
	 * @param aId {integer}
   * @return {extensions.tbnotes.data.NotePrototype|NULL}
	 */
	this.getNote = function(aId) {
		return this.getNoteById();
	}
	
	// Alias for getNote
	this.getNoteById = function(aId) 
  {
		var query = "SELECT * " + 
		            "FROM tbn_notes " + 
                "WHERE id = :note_id " + 
                "LIMIT 1 ";
		var stmt = __dbCon.createStatement(query);
		stmt.params.note_id = aId;
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
        return new extensions.tbnotes.data.NotePrototype(stmt.row);
			}
		}
		
		return null;
	};
	
	/**
	 * Returns note which is attached to an email message with given key. 
	 * Otherwise returns NULL.
	 * 
	 * @param aKey {string}
   * @return {extensions.tbnotes.data.NotePrototype|NULL}
	 */
	this.getNoteByTargetKey= function(aKey) 
  {
		var query = "SELECT * " + 
		            "FROM tbn_notes " + 
                "WHERE " + 
                "  target_key = :target_key AND " + 
                "  deleted = 0 " + 
                "LIMIT 1 ";
		var stmt = __dbCon.createStatement(query);
		stmt.params.target_key = aKey;
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
        return new extensions.tbnotes.data.NotePrototype(stmt.row);
			}
		}
		
		return null;
	};
  
  /** 
   * Returns all notes as objects.
   * 
   * @param aAlsoDeleted {boolean}
   * @return {array} Array of instances of extensions.tbnotes.data.NotePrototype
   */
  this.getNotes = function(aAlsoDeleted)
  {
    var notes = new Array();
    var query = "SELECT * FROM tbn_notes " + 
        ((aAlsoDeleted !== true) ? "WHERE deleted = 0 " : "");
    var stmt = __dbCon.createStatement(query);
    
    while(stmt.executeStep()) {
      if(stmt.row.id > 0) {
        notes.push((new extensions.tbnotes.data.NotePrototype(stmt.row)));
      }
    }
    
    return notes;
  };
	
	/**
	 * Adds new note into the database
	 * 
	 * @param aNote {extensions.tbnotes.data.NotePrototype}
	 * @return {Integer|NULL}
	 */
	this.addNote = function(aNote) 
  {
		try {
			var query = 
					"INSERT INTO tbn_notes " + 
					"  (target_type, target_key, target_name, category_id, created, updated, title, body, deleted) " + 
					"VALUES " + 
					"  ( :target_type , :target_key , :target_name , 0 , :created , :updated , :title , :body , 0); ";
			
			var stmt = __dbCon.createStatement(query);
			stmt.params.target_type = aNote.mTargetType;
			stmt.params.target_key  = aNote.mTargetKey;
			stmt.params.target_name = aNote.mTargetName;
			stmt.params.created = extensions.tbnotes.getCurrentDatetime();
			stmt.params.updated = extensions.tbnotes.getCurrentDatetime();
			stmt.params.title = aNote.mTitle;
			stmt.params.body = aNote.mBody;
			stmt.execute();
		} catch(e) {
			CU.reportError("Database error: [" + __dbCon.lastError + "]: " + __dbCon.lastErrorString);
      return null;
		}
    
    return __dbCon.lastInsertRowID;
	};
  
  /**
   * Removes note with given ID permanently from the database.
   * 
   * @param aId {Integer}
   * @return {Boolean}
   */
  this.removeNote = function(aId) 
  {
    try {
			var query = "DELETE FROM tbn_notes WHERE id = :note_id LIMIT 1 ";
			var stmt = __dbCon.createStatement(query);
			stmt.params.note_id = aId;
			stmt.execute();
    } catch(e) {
			CU.reportError("Database error: [" + __dbCon.lastError + "]: " + __dbCon.lastErrorString);
			return false;
    }
    
    return true;
  };
	
	/**
	 * Updates note's record in the database
	 * 
	 * @param aNote {extensions.tbnotes.data.NotePrototype}
	 * @return {Boolean}
	 */
	this.updateNote = function(aNote) 
  {
		try {
			// FIXME This query doesn't working!
			var query = 
        "UPDATE tbn_notes " + 
        "SET " + 
        "  target_type = :target_type , " + 
        "  target_key = :target_key , " + 
        "  target_name = :target_name , " + 
        "  updated = :updated , " + 
        "  title = :title ,  " + 
        "  body = :body , " +
				"  deleted = :deleted " + 
        "WHERE " + 
        "  id = :note_id ";
			
			var stmt = __dbCon.createStatement(query);
			stmt.params.target_type = aNote.mTargetType;
			stmt.params.target_key = aNote.mTargetKey;
			stmt.params.target_name = aNote.mTargetName;
			stmt.params.updated = extensions.tbnotes.getCurrentDatetime();
			stmt.params.title = aNote.mTitle;
			stmt.params.body = aNote.mBody;
			stmt.params.note_id = aNote.mId;
			stmt.params.deleted = (aNote.isMarkedAsDeleted()) ? 1 : 0;
			stmt.execute();
		} catch(e) {
			CU.reportError("Database error: [" + __dbCon.lastError + "]: " + __dbCon.lastErrorString);
      return false;
		}
    
    return true;
	};
  
  /**
   * Mark note with given ID as deleted.
   * 
   * @param aId {Integer}
   * @return {Boolean}
   */
  this.deleteNote = function(aId) 
  {
		try {
			var query = "UPDATE tbn_notes SET deleted = 1 WHERE id = :note_id ";
			var stmt = __dbCon.createStatement(query);
			stmt.params.note_id = aId;
			stmt.execute();
		} catch(e) {
			CU.reportError("Database error: [" + __dbCon.lastError + "]: " + __dbCon.lastErrorString);
			return false;
		}
    
    return true;
	};
  
  // ========================================================================
  // Categories
  
	// TODO Implement `getCategory` method!
	// TODO Implement `getCategories` method!
  // TODO Implement `addCategory` method!
  // TODO Implement `removeCategory` method!
  // TODO Implement `updateCategory` method!
  
  // ========================================================================
  // Links
	
	/**
	 * Returns link by its id. Otherwise returns NULL.
	 * 
	 * @param aId {integer}
   * @return {extensions.tbnotes.data.LinkPrototype|NULL}
	 */
	this.getLink = function(aId) {
		return this.getLinkById(aId);
	};
	
	// Alias for getLink
	this.getLinkById = function(aId) 
  {
		var query = "SELECT * FROM tbn_links WHERE id = :link_id ";
		var stmt = __dbCon.createStatement(query);
		stmt.params.link_id = aId;
		
		while(stmt.executeStep()) {
			if(stmt.row.id > 0) {
        return new extensions.tbnotes.data.LinkPrototype(stmt.row);
			}
		}
		
		return null;
	};
  
  /** 
   * Returns all links as objects.
   * 
   * @return {array} Array of instances of extensions.tbnotes.data.LinkPrototype
   */
  this.getLinks = function()
  {
    var links = new Array();
    var stmt = __dbCon.createStatement("SELECT * FROM tbn_links ");
    
    while(stmt.executeStep()) {
      if(stmt.row.id > 0) {
        links.push((new extensions.tbnotes.data.LinkPrototype(stmt.row)));
      }
    }
    
    return links;
  };
  
  /** 
   * Returns all links which are attached to the note with given ID.
   *
   * @param aNoteId {integer}
   * @return {array} Array of instances of extensions.tbnotes.data.LinkPrototype
   */
  this.getLinksByNoteId = function(aNoteId)
  {
    var links = new Array();
    var stmt = __dbCon.createStatement("SELECT * FROM tbn_links WHERE note_id = '" + aNoteId + "' ");
    
    while(stmt.executeStep()) {
      if(stmt.row.id > 0) {
        links.push((new extensions.tbnotes.data.LinkPrototype(stmt.row)));
      }
    }
    
    return links;
  };
	
	/**
	 * Adds new link into the database
	 * 
	 * @param aLink {extensions.tbnotes.data.LinkPrototype}
	 * @return {Integer|NULL}
	 */
	this.addLink = function(aLink) 
  {
		try {
			var query = 
					"INSERT INTO tbn_links (note_id, url, title) " + 
					"VALUES ( :note_id , :url , :title ); ";
			
			var stmt = __dbCon.createStatement(query);
			stmt.params.note_id = aLink.mNoteId;
			stmt.params.url = aLink.mUrl;
			stmt.params.title = aLink.mTitle;
			stmt.execute();
		} catch(e) {
			CU.reportError("Database error: [" + __dbCon.lastError + "]: " + __dbCon.lastErrorString);
      return null;
		}
    
    return __dbCon.lastInsertRowID;
	};
  
  /**
   * Removes link with given ID permanently from the database.
   * 
   * @param aId {Integer}
   * @return {Boolean}
   */
  this.removeLink = function(aId) 
  {
    try {
			var query = "DELETE FROM tbn_links WHERE id = :link_id ";
			var stmt = __dbCon.createStatement(query);
			stmt.params.link_id = aId;
			stmt.execute();
    } catch(e) {
			CU.reportError("Database error: [" + __dbCon.lastError + "]: " + __dbCon.lastErrorString);
			return false;
    }
    
    return true;
  };
	
	/**
	 * Updates link
	 * 
	 * @param aLink {extensions.tbnotes.data.LinkPrototype}
	 * @return {Boolean}
	 */
	this.updateLink = function(aLink) 
  {
		try {
			var query = 
        "UPDATE tbn_links " + 
        "SET note_id = :note_id , url = :url , title = :title " + 
        "WHERE id = :link_id ";
			
			var stmt = __dbCon.createStatement(query);
			stmt.params.note_id = aLink.mNoteId;
			stmt.params.url = aLink.mUrl;
			stmt.params.title = aLink.mTitle;
			stmt.params.link_id = aLink.mId;
			stmt.execute();
		} catch(e) {
			CU.reportError("Database error: [" + __dbCon.lastError + "]: " + __dbCon.lastErrorString);
      return false;
		}
    
    return true;
	};

}).apply(extensions.tbnotes.db);
